import java.util.Arrays;
import java.util.List;
import org.virtualbox_5_0.CleanupMode;
import org.virtualbox_5_0.IEvent;
import org.virtualbox_5_0.IEventListener;
import org.virtualbox_5_0.IMachine;
import org.virtualbox_5_0.IMedium;
import org.virtualbox_5_0.IProgress;
import org.virtualbox_5_0.ISession;
import org.virtualbox_5_0.IVirtualBox;
import org.virtualbox_5_0.SessionState;
import org.virtualbox_5_0.VBoxEventType;
import org.virtualbox_5_0.VBoxException;
import org.virtualbox_5_0.VirtualBoxManager;

/**
 * Test class for <a href="https://www.virtualbox.org/ticket/13647">Ticket #13647</a>
 * <p>
 * Updated to take into account <code>VirtualBoxManager::waitForEvent()</code>
 * </p>
 * 
 * @author Maxime Dor
 */
public class OutOfMemory_5_0_RC2 extends Thread {

   private static long machineId = 0L;
   private static final int waitingCoef = 500;
   private static Thread worker;
   private static VirtualBoxManager vboxManager;

   public synchronized static String getNextName() {
      return Long.toString(machineId++);
   }

   public static void main(String[] args) {
      int nbThreads = Integer.parseInt(System.getProperty("oom.thread",Integer.toString(Runtime.getRuntime().availableProcessors())));

      Runtime.getRuntime().addShutdownHook(new Thread() {

         @Override
         public void run() {
            System.out.println("Disconnecting");
            worker.interrupt();
            try {
               worker.join();

            } catch (InterruptedException e) {
               e.printStackTrace();
            }
            vboxManager.cleanup();
            System.out.println("Disconnected");
         }
      });

      System.out.println("Starting");
      vboxManager = VirtualBoxManager.createInstance(null);
      if (System.getProperty("vbox.ws") != null) {
         vboxManager.connect("http://" + System.getProperty("vbox.ws.host", "localhost") + ":" + System.getProperty("vbox.ws.port", "18083"), "", "");
      }
      System.out.println("Connected");
      IVirtualBox vbox = vboxManager.getVBox();

      worker = new EventWorker(vbox);
      worker.start();

      for (int i = 0; i < nbThreads; i++) {
         (new OutOfMemory_5_0_RC2()).start();
      }

   }

   @Override
   public void run() {
      while (worker.isAlive()) {
         try {

            IMachine machine = vboxManager.getVBox().createMachine(null, getNextName(), null, "Other", null);
            vboxManager.getVBox().registerMachine(machine);
            System.out.println("Machine registered");

            ISession session = vboxManager.getSessionObject();
            IProgress pOn = machine.launchVMProcess(session, "headless", null);
            while (!pOn.getCompleted() || pOn.getCanceled()) {
               try {
                  Thread.sleep(Math.abs(pOn.getTimeRemaining()) * waitingCoef);
               } catch (InterruptedException e) {
                  e.printStackTrace();
               }
            }
            if (pOn.getResultCode() != 0) {
               System.out.println("Error when starting machine: " + pOn.getErrorInfo().getText());
            } else {
               System.out.println("Maching started");
            }

            IProgress pOff = session.getConsole().powerDown();
            while (!pOff.getCompleted() || pOff.getCanceled()) {
               try {
                  Thread.sleep(Math.abs(pOn.getTimeRemaining()) * waitingCoef);
               } catch (InterruptedException e) {
                  e.printStackTrace();
               }
            }
            if (pOff.getResultCode() != 0) {
               System.out.println("Error when stopping machine: " + pOff.getErrorInfo().getText());
            } else {
               System.out.println("Machine stopped");
            }

            if (session.getState().equals(SessionState.Locked)) {
               try {
                  session.unlockMachine();
               } catch (VBoxException e) {
                  if (!e.getMessage().toUpperCase().contains("8000FFFF")) {
                     e.printStackTrace();
                  }

               }
            }

            machine = vboxManager.getVBox().findMachine(machine.getId());
            while (!machine.getSessionState().equals(SessionState.Unlocked)) {
               Thread.sleep(10);
            }
            try {
               List<IMedium> hdds = machine.unregister(CleanupMode.DetachAllReturnHardDisksOnly);
               IProgress pDel = machine.deleteConfig(hdds);
               while (!pDel.getCompleted() || pDel.getCanceled()) {
                  try {
                     Thread.sleep(Math.abs(pDel.getTimeRemaining()) * waitingCoef);
                  } catch (InterruptedException e) {
                     e.printStackTrace();
                  }
               }
               if (pDel.getResultCode() != 0) {
                  System.out.println("Error when deleting machine: " + pDel.getErrorInfo().getText());
               } else {
                  System.out.println("Machine deleted");
               }
            } catch (Throwable t) {
               t.printStackTrace();
            }

            System.out.println("------------------------------");
         } catch (Throwable t) {
            t.printStackTrace();
            worker.interrupt();
            break;
         }
      }

      System.out.println("Stopping");
   }

   private static class EventWorker extends Thread {

      private IVirtualBox vbox;

      public EventWorker(IVirtualBox vbox) {
         this.vbox = vbox;
      }

      @Override
      public void run() {
         System.out.println("Starting worker");
         IEventListener el = vbox.getEventSource().createListener();
         vbox.getEventSource().registerListener(el, Arrays.asList(VBoxEventType.Any), false);

         while (!Thread.currentThread().isInterrupted()) {
            try {
               IEvent rawEvent = vbox.getEventSource().getEvent(el, 1000);
               if (rawEvent != null) {
                  System.out.println("VBox Event: " + rawEvent.getClass().getName() + " - " + rawEvent.getType() + " - " + rawEvent);
                  vbox.getEventSource().eventProcessed(el, rawEvent);
               }
               vboxManager.waitForEvents(0);
            } catch (Throwable t) {
               t.printStackTrace();
               break;
            }
         }

         vbox.getEventSource().unregisterListener(el);
         System.out.println("Stopping worker");
      }

   }

}
